/* -*- buffer-read-only: t -*- vi: set ro:
   THIS FILE IS GENERATED AUTOMATICALLY.  PLEASE DO NOT EDIT.
*/
#define MF_SOURCE_NAME  MF_SOURCE_BI_DB
#ifdef HAVE_CONFIG_H
# include <config.h>
#endif
#include <sys/types.h>
#include "mailfromd.h"
#include "prog.h"
#include "builtin.h"

#line 1 "bi_db.m4"
/* This file is part of Mailfromd.             -*- c -*-
   Copyright (C) 2006, 2007, 2008, 2009 Sergey Poznyakoff

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>. */

#define DEFAULT_DB_MODE 0640

struct db_prop {                /* Database properties */
	char *pat;              /* Database name pattern */
	mode_t mode;            /* File mode */
	int null;               /* Null byte */
};

static mu_list_t db_prop_list;

static int
db_prop_compare(const void *item, const void *data)
{
	const struct db_prop *prop = item;
	const char *name = data;
	return strcmp(prop->pat, name);
}

static int
strtomode(char *str, mode_t *pmode)
{
	mode_t mode = 0;
	struct { char c; unsigned mask; } modetab[] = {
		{ 'r', S_IRUSR },
		{ 'w', S_IWUSR },
		{ 'x', S_IXUSR },
		{ 'r', S_IRGRP },
		{ 'w', S_IWGRP },
		{ 'x', S_IXGRP },
		{ 'r', S_IROTH },
		{ 'w', S_IWOTH },
		{ 'x', S_IXOTH },
		{ 0 }
	};
	int i;

	for (i = 0; modetab[i].c; i++) {
		if (!str[i])
			return i + 1;
		if (str[i] == modetab[i].c)
			mode |= modetab[i].mask;
		else if (str[i] != '-')
			return i + 1;
	}
	if (str[i])
		return i + 1;
	*pmode = mode;
	return 0;
}

/* #pragma dbprop <name> [null] [mode] */
#line 68 "bi_db.m4"

#line 68
static void _pragma_dbprop (int argc, char **argv, const char *text)
#line 68

{
	int null = 0;
	mode_t mode = DEFAULT_DB_MODE;
	struct db_prop *prop;
	int rc;
	char *pat;

	--argc;
	pat = *++argv;
	
	while (--argc) {
		char *p = *++argv;

		if (strcmp (p, "null") == 0)
			null = 1;
		else if (c_isdigit(*p)) {
			unsigned long n = strtoul(p, &p, 8);
			if (*p || (mode = n) != n) {
				parse_error(_("bad numeric file mode"));
				return;
			}
		} else if (rc = strtomode(p, &mode)) {
			parse_error(_("bad symbolic file mode (near %s)"),
				    p + rc - 1);
			return;
		}
	}

	if (!db_prop_list) {
		rc = mu_list_create(&db_prop_list);
		if (rc) {
			parse_error(_("Cannot create list: %s"),
				    mu_strerror(rc));
			return;
		}
		mu_list_set_comparator(db_prop_list, db_prop_compare);
	}

	if (mu_list_locate(db_prop_list, pat, (void**) &prop)) {
		prop = xmalloc(sizeof(*prop));
		prop->pat = xstrdup(pat);
		rc = mu_list_append(db_prop_list, prop);
		if (rc) {
			parse_error(_("Cannot create list: %s"),
				    mu_strerror(rc));
			return;
		}
	}
	prop->mode = mode;
	prop->null = null;
}

const struct db_prop *
db_prop_lookup(const char *name)
{
	mu_iterator_t itr;
	const struct db_prop *found = NULL;
	
	if (db_prop_list && mu_list_get_iterator(db_prop_list, &itr)) {
		for (mu_iterator_first(itr);
		     !mu_iterator_is_done(itr);
		     mu_iterator_next(itr)) {
			const struct db_prop *p;
			mu_iterator_current(itr, (void**)&p);
			if (strcmp(p->pat, name) == 0) {
				found = p;
				break;
			} else if (fnmatch(p->pat, name, 0) == 0)
				found = p;
		}
		mu_iterator_destroy(&itr);
	}
	return found;	
}


#define LOOKUP_NULL_BYTE 0x1
#define LOOKUP_TEST_ONLY 0x2

static int
dbmap_lookup(eval_environ_t env, char *dbname, const char *keystr,
	     const char *defval, int flags)
{
	int rc;
	DBM_FILE db;
	DBM_DATUM key;
	DBM_DATUM contents;

	if (!defval)
		defval = "";
	if (mu_dbm_open(dbname, &db, MU_STREAM_READ, 0, NULL))
			env_throw_bi(env, mfe_dbfailure, NULL, _("mu_dbm_open(%s) failed: %s"),dbname,mu_dbm_strerror());
#line 164

	memset(&key, 0, sizeof key);
	memset(&contents, 0, sizeof contents);
	MU_DATUM_PTR(key) = (void*) keystr;
	MU_DATUM_SIZE(key) = strlen(keystr);
	if (flags & LOOKUP_NULL_BYTE)
		MU_DATUM_SIZE(key)++;
	rc = mu_dbm_fetch(&db, key, &contents) == 0;
	debug2(10, "Looking up %s: %s", keystr, rc ? "true" : "false");
	if (flags & LOOKUP_TEST_ONLY) 
		push(env, (STKVAL)rc);
	else {
		if (!rc) {
			if (defval)
				pushs(env, (STKVAL)defval);
			else
				push(env, 0);
		} else if (((char*)MU_DATUM_PTR(contents))[MU_DATUM_SIZE(contents)-1]) {
			size_t off;
			size_t len = MU_DATUM_SIZE(contents);
			char *s = (char*) env_data_ref(env, (off = heap_reserve(env, len + 1)));
			memcpy(s, MU_DATUM_PTR(contents), len);
			s[len] = 0;
			push(env, (STKVAL) off);
		} else
			pushs(env, MU_DATUM_PTR(contents));
	}
	mu_dbm_datum_free(&contents);
	mu_dbm_close(&db);
	return rc;
}

void
#line 196
bi_dbmap(eval_environ_t env)
#line 196

#line 196 "bi_db.m4"
{
#line 196
	
#line 196

#line 196
long __bi_argcnt;
#line 196
char * dbname;
#line 196
        char * key;
#line 196
        long  null;
#line 196
        
#line 196
get_string_arg(env, 1, &dbname);
#line 196
        get_string_arg(env, 2, &key);
#line 196
        get_numeric_arg(env, 3, &null);
#line 196
        
#line 196
get_numeric_arg(env, 0, &__bi_argcnt);
#line 196
        adjust_stack(env, __bi_argcnt + 1);
#line 196

#line 196
	if (prog_trace_option & BUILTIN_MASK_db)
#line 196
		prog_trace(env, "dbmap %s %s %lu",dbname, key, ((__bi_argcnt > 2) ? null : 0));;
#line 196

{
	const struct db_prop *prop = db_prop_lookup(dbname);
	dbmap_lookup(env, dbname, key, NULL,
		     LOOKUP_TEST_ONLY | 
		         (((__bi_argcnt > 2) ? null : prop && prop->null)
			   ? LOOKUP_NULL_BYTE : 0));
}
}

void
#line 206
bi_dbget(eval_environ_t env)
#line 206

#line 206 "bi_db.m4"
{
#line 206
	
#line 206

#line 206
long __bi_argcnt;
#line 206
char * dbname;
#line 206
        char * key;
#line 206
        char * defval;
#line 206
        long  null;
#line 206
        
#line 206
get_string_arg(env, 1, &dbname);
#line 206
        get_string_arg(env, 2, &key);
#line 206
        get_string_arg(env, 3, &defval);
#line 206
        get_numeric_arg(env, 4, &null);
#line 206
        
#line 206
get_numeric_arg(env, 0, &__bi_argcnt);
#line 206
        adjust_stack(env, __bi_argcnt + 1);
#line 206

#line 206
	if (prog_trace_option & BUILTIN_MASK_db)
#line 206
		prog_trace(env, "dbget %s %s %s %lu",dbname, key, ((__bi_argcnt > 2) ? defval : ""), ((__bi_argcnt > 3) ? null : 0));;

{
	const struct db_prop *prop = db_prop_lookup(dbname);
	dbmap_lookup(env, dbname, key, 
                     ((__bi_argcnt > 2) ? defval : 0), 
		     ((__bi_argcnt > 3) ? null : prop && prop->null)
		      ? LOOKUP_NULL_BYTE : 0);
}
}

void
#line 217
bi_dbput(eval_environ_t env)
#line 217

#line 217 "bi_db.m4"
{
#line 217
	
#line 217

#line 217
long __bi_argcnt;
#line 217
char * dbname;
#line 217
        char * keystr;
#line 217
        char * value;
#line 217
        long  null;
#line 217
        long  mode;
#line 217
        
#line 217
get_string_arg(env, 1, &dbname);
#line 217
        get_string_arg(env, 2, &keystr);
#line 217
        get_string_arg(env, 3, &value);
#line 217
        get_numeric_arg(env, 4, &null);
#line 217
        get_numeric_arg(env, 5, &mode);
#line 217
        
#line 217
get_numeric_arg(env, 0, &__bi_argcnt);
#line 217
        adjust_stack(env, __bi_argcnt + 1);
#line 217

#line 217
	if (prog_trace_option & BUILTIN_MASK_db)
#line 217
		prog_trace(env, "dbput %s %s %s %lu %lu",dbname, keystr, value, ((__bi_argcnt > 3) ? null : 0), ((__bi_argcnt > 4) ? mode : 0));;

{
	int rc;
	DBM_FILE db;
	DBM_DATUM key;
	DBM_DATUM contents;
	const struct db_prop *prop = db_prop_lookup(dbname);
	
	if (mu_dbm_open(dbname, &db, MU_STREAM_RDWR,
			((__bi_argcnt > 4) ? mode : (prop ? prop->mode : DEFAULT_DB_MODE)),
			NULL))
			env_throw_bi(env, mfe_dbfailure, "dbput", _("mu_dbm_open(%s) failed: %s"),dbname,mu_dbm_strerror());
#line 233
	memset(&key, 0, sizeof key);
	MU_DATUM_PTR(key) = keystr;
	MU_DATUM_SIZE(key) = strlen(keystr);
	if (((__bi_argcnt > 3) ? null : prop && prop->null))
		MU_DATUM_SIZE(key)++;

	memset(&contents, 0, sizeof contents);
	MU_DATUM_PTR(contents) = value;
	MU_DATUM_SIZE(contents) = strlen(value) + 1;
	
	rc = mu_dbm_insert(&db, key, contents, 1);
	mu_dbm_close(&db);
		if (!(rc == 0))
#line 245
			env_throw_bi(env, mfe_dbfailure, "dbput", _("Failed to insert data to %s: %s %s: %s"),dbname,keystr,value,mu_dbm_strerror())
#line 251
;
}
}

void
#line 255
bi_dbdel(eval_environ_t env)
#line 255

#line 255 "bi_db.m4"
{
#line 255
	
#line 255

#line 255
long __bi_argcnt;
#line 255
char * dbname;
#line 255
        char * keystr;
#line 255
        long  null;
#line 255
        long  mode;
#line 255
        
#line 255
get_string_arg(env, 1, &dbname);
#line 255
        get_string_arg(env, 2, &keystr);
#line 255
        get_numeric_arg(env, 3, &null);
#line 255
        get_numeric_arg(env, 4, &mode);
#line 255
        
#line 255
get_numeric_arg(env, 0, &__bi_argcnt);
#line 255
        adjust_stack(env, __bi_argcnt + 1);
#line 255

#line 255
	if (prog_trace_option & BUILTIN_MASK_db)
#line 255
		prog_trace(env, "dbdel %s %s %lu %lu",dbname, keystr, ((__bi_argcnt > 2) ? null : 0), ((__bi_argcnt > 3) ? mode : 0));;

{
	DBM_FILE db;
	DBM_DATUM key;
	int rc;
	const struct db_prop *prop = db_prop_lookup(dbname);
	
	if (mu_dbm_open(dbname, &db, MU_STREAM_RDWR,
			((__bi_argcnt > 3) ? mode : (prop ? prop->mode : DEFAULT_DB_MODE)),
			NULL))
			env_throw_bi(env, mfe_dbfailure, "dbdel", _("mu_dbm_open(%s) failed: %s"),dbname,mu_dbm_strerror());
#line 270
	memset(&key, 0, sizeof key);
	MU_DATUM_PTR(key) = keystr;
	MU_DATUM_SIZE(key) = strlen(keystr);
	if (((__bi_argcnt > 2) ? null : prop && prop->null))
		MU_DATUM_SIZE(key)++;
	rc = mu_dbm_delete(&db, key);
	mu_dbm_close(&db);
		if (!(rc == 0))
#line 277
			env_throw_bi(env, mfe_dbfailure, "dbdel", _("Failed to delete data `%s' from `%s': %s"),keystr,dbname,mu_dbm_strerror())
#line 282
;
}
}


#define NUMDB 128

struct db_tab {
	int used;
	DBM_FILE db;
	DBM_DATUM key;
};

static void *
alloc_db_tab()
{
	return xcalloc(NUMDB, sizeof(struct db_tab));
}

static void
close_db_tab(struct db_tab *dbt)
{
	if (dbt->used) {
		mu_dbm_datum_free(&dbt->key);
		mu_dbm_close(&dbt->db);
		dbt->used = 0;
	}
}

static void
destroy_db_tab(void *data)
{
	int i;
	struct db_tab *db = data;
	for (i = 0; i < NUMDB; i++) 
		close_db_tab(db + i);
	free(db);
}


#line 321

#line 321
static int DBTAB_id;
#line 321 "bi_db.m4"
;

static int
new_db_tab(struct db_tab *dbt)
{
	int i;
	for (i = 0; i < NUMDB; i++)
		if (!dbt[i].used) {
			dbt[i].used = 1;
			return i;
		}
	return -1;
}
			


void
#line 337
bi_dbfirst(eval_environ_t env)
#line 337

#line 337 "bi_db.m4"
{
#line 337
	
#line 337

#line 337

#line 337
char * dbname;
#line 337
        
#line 337
get_string_arg(env, 0, &dbname);
#line 337
        
#line 337

#line 337
        adjust_stack(env, 1);
#line 337

#line 337
	if (prog_trace_option & BUILTIN_MASK_db)
#line 337
		prog_trace(env, "dbfirst %s",dbname);;
#line 337

{
	int rc;
	int n;
	struct db_tab *dbt = env_get_builtin_priv(env,DBTAB_id);
	DBM_FILE db;
	DBM_DATUM key;
	
	if (mu_dbm_open(dbname, &db, MU_STREAM_READ, 0, NULL))
			env_throw_bi(env, mfe_dbfailure, "dbfirst", _("mu_dbm_open(%s) failed: %s"),dbname,mu_dbm_strerror());
#line 350
	rc = mu_dbm_firstkey(&db, &key);
		if (!(rc == 0))
#line 351
			env_throw_bi(env, mfe_dbfailure, "dbfirst", _("mu_dbm_firstkey failed: %s"),mu_dbm_strerror())
#line 353
;
	n = new_db_tab(dbt);
		if (!(n >= 0))
#line 355
			env_throw_bi(env, mfe_failure, "dbfirst", _("No more database entries available"))
#line 357
;
	dbt += n;
	dbt->db = db;
	dbt->key = key;
	
#line 361
do {
#line 361
  push(env, (STKVAL) (n));
#line 361
  return;
#line 361
} while (0);
}
}

void
#line 365
bi_dbnext(eval_environ_t env)
#line 365

#line 365 "bi_db.m4"
{
#line 365
	
#line 365

#line 365

#line 365
long  dn;
#line 365
        
#line 365
get_numeric_arg(env, 0, &dn);
#line 365
        
#line 365

#line 365
        adjust_stack(env, 1);
#line 365

#line 365
	if (prog_trace_option & BUILTIN_MASK_db)
#line 365
		prog_trace(env, "dbnext %lu",dn);;
#line 365

{
	struct db_tab *dbt = env_get_builtin_priv(env,DBTAB_id) + dn;
	DBM_DATUM nextkey;
	int rc;
	
		if (!(dn >= 0 && dn < NUMDB && dbt->used))
#line 371
			env_throw_bi(env, mfe_range, "dbnext", _("Invalid database descriptor"))
#line 373
;

	rc = mu_dbm_nextkey(&dbt->db, dbt->key, &nextkey);
	if (rc) {
		close_db_tab(dbt);
		
#line 378
do {
#line 378
  push(env, (STKVAL) (0));
#line 378
  return;
#line 378
} while (0);
	}
	mu_dbm_datum_free(&dbt->key);
	dbt->key = nextkey;
	
#line 382
do {
#line 382
  push(env, (STKVAL) (1));
#line 382
  return;
#line 382
} while (0);
}
}

void
#line 386
bi_dbkey(eval_environ_t env)
#line 386

#line 386 "bi_db.m4"
{
#line 386
	
#line 386

#line 386

#line 386
long  dn;
#line 386
        
#line 386
get_numeric_arg(env, 0, &dn);
#line 386
        
#line 386

#line 386
        adjust_stack(env, 1);
#line 386

#line 386
	if (prog_trace_option & BUILTIN_MASK_db)
#line 386
		prog_trace(env, "dbkey %lu",dn);;
#line 386

{
	size_t off, len;
	char *s;
	struct db_tab *dbt = env_get_builtin_priv(env,DBTAB_id) + dn;
	
		if (!(dn >= 0 && dn < NUMDB && dbt->used))
#line 392
			env_throw_bi(env, mfe_range, "dbkey", _("Invalid database descriptor"))
#line 394
;
	
	len = MU_DATUM_SIZE(dbt->key);
	s = (char*) env_data_ref(env, (off = heap_reserve(env, len + 1)));
	memcpy(s, MU_DATUM_PTR(dbt->key), len);
	s[len] = 0;
	
#line 400
do {
#line 400
  push(env, (STKVAL) (off));
#line 400
  return;
#line 400
} while (0);
}
}

void
#line 404
bi_dbvalue(eval_environ_t env)
#line 404

#line 404 "bi_db.m4"
{
#line 404
	
#line 404

#line 404

#line 404
long  dn;
#line 404
        
#line 404
get_numeric_arg(env, 0, &dn);
#line 404
        
#line 404

#line 404
        adjust_stack(env, 1);
#line 404

#line 404
	if (prog_trace_option & BUILTIN_MASK_db)
#line 404
		prog_trace(env, "dbvalue %lu",dn);;
#line 404

{
	int rc;
	size_t off, len;
	char *s;
	struct db_tab *dbt = env_get_builtin_priv(env,DBTAB_id) + dn;
	DBM_DATUM contents;
	
		if (!(dn >= 0 && dn < NUMDB && dbt->used))
#line 412
			env_throw_bi(env, mfe_range, "dbvalue", _("Invalid database descriptor"))
#line 414
;

	memset(&contents, 0, sizeof contents);
	rc = mu_dbm_fetch(&dbt->db, dbt->key, &contents);
		if (!(rc == 0))
#line 418
			env_throw_bi(env, mfe_dbfailure, "dbvalue", _("Key not found: %s"),mu_dbm_strerror())
#line 421
;
	
	len = MU_DATUM_SIZE(contents);
	s = (char*) env_data_ref(env, (off = heap_reserve(env, len + 1)));
	memcpy(s, MU_DATUM_PTR(contents), len);
	s[len] = 0;
	mu_dbm_datum_free(&contents);
	
#line 428
do {
#line 428
  push(env, (STKVAL) (off));
#line 428
  return;
#line 428
} while (0);
}
}


static void
greylist_print_item(const char *key, size_t size, const void *content)
{
	time_t timestamp = *(time_t*) content;
	size--; /* Size includes the trailing nul */
	printf("%*.*s ", size, size, key);
	format_time_str(stdout, timestamp);
	putchar('\n');
}

static int
greylist_expire_item(const void *content)
{
	time_t timestamp = *(time_t*) content;
	return greylist_format->expire_interval &&
		time(NULL) - timestamp > greylist_format->expire_interval;
}


static struct db_format greylist_format_struct = {
	"greylist",
	DEFAULT_GREYLIST_DATABASE,
	1,
	DEFAULT_EXPIRE_INTERVAL,
	greylist_print_item,
	greylist_expire_item
};

struct db_format *greylist_format = &greylist_format_struct;

static size_t greylist_seconds_left_loc
#line 463 "bi_db.m4"
;

/* greylist(key, interval)

   Returns true if the key is greylisted, false if it's OK to
   deliver mail.
 */
void
#line 470
bi_greylist(eval_environ_t env)
#line 470

#line 470 "bi_db.m4"
{
#line 470
	
#line 470

#line 470

#line 470
char * email;
#line 470
        long  interval;
#line 470
        
#line 470
get_string_arg(env, 0, &email);
#line 470
        get_numeric_arg(env, 1, &interval);
#line 470
        
#line 470

#line 470
        adjust_stack(env, 2);
#line 470

#line 470
	if (prog_trace_option & BUILTIN_MASK_db)
#line 470
		prog_trace(env, "greylist %s %lu",email, interval);;
#line 470

{
	int rc;
	DBM_FILE db;
	DBM_DATUM key;
	DBM_DATUM contents;
	int readonly;
	time_t now;

	rc = mu_dbm_open(greylist_format->dbname, &db, MU_STREAM_RDWR, 0600,
			 &readonly);
        	if (!(rc == 0))
#line 481
			env_throw_bi(env, mfe_dbfailure, "greylist", _("mu_dbm_open(%s) failed: %s"),greylist_format->dbname,mu_dbm_strerror())
;
	
	memset(&key, 0, sizeof key);
	memset(&contents, 0, sizeof contents);
	MU_DATUM_PTR(key) = email;
	MU_DATUM_SIZE(key) = strlen(email)+1;

	time(&now);
	if (mu_dbm_fetch(&db, key, &contents) == 0) {
		time_t timestamp, diff;

			if (!(MU_DATUM_SIZE(contents) == sizeof timestamp))
#line 493
			env_throw_bi(env, mfe_dbfailure, "greylist", _("Greylist database %s has wrong data size"),greylist_format->dbname)
#line 496
;
		
		timestamp = *(time_t*) MU_DATUM_PTR(contents);
		diff = now - timestamp;
	
		__DBG(20) {
			char timebuf[32];
			debug_log("%s entered greylist database on %s, "
				  "%ld seconds ago",
				  email,
				  mailfromd_timestr(timestamp, timebuf,
						    sizeof timebuf),
				  (long) diff);
		}

		if (diff < interval) {
			diff = interval - diff;

			*env_data_ref(env, greylist_seconds_left_loc) = (STKVAL) (diff);
			
			debug2(20, "%s still greylisted (for %lu sec.)",
			       email,
			       (unsigned long) diff);
			rc = 1;
		} else if (diff > greylist_format->expire_interval) {
			debug1(20, "greylist record for %s expired",
			       email);
			if (!readonly) {
				*env_data_ref(env, greylist_seconds_left_loc) = (STKVAL) (interval);
				memcpy(MU_DATUM_PTR(contents),
				       &now, sizeof now);
				if (mu_dbm_insert(&db, key, contents, 1))
					mu_error(_("Cannot insert datum `%-.*s' in "
						   "greylist database %s: %s"),
			                         MU_DATUM_SIZE(key),
						 (char*)MU_DATUM_PTR(key),
						 greylist_format->dbname,
	                                         mu_dbm_strerror());
			} else
				debug(20, "database opened in readonly mode: "
				      "not updating");
			rc = 1;
		} else {
			debug1(20, "%s finished greylisting period",
			       email);
			rc = 0;
		}
		mu_dbm_datum_free(&contents);
	} else if (!readonly) {
		debug1(20, "greylisting %s", email);
		*env_data_ref(env, greylist_seconds_left_loc) = (STKVAL) (interval);
		MU_DATUM_PTR(contents) = (void*)&now;
		MU_DATUM_SIZE(contents) = sizeof now;
		if (mu_dbm_insert(&db, key, contents, 1))
			mu_error(_("Cannot insert datum `%-.*s' in greylist "
				   "database %s: %s"),
		                 MU_DATUM_SIZE(key), (char*)MU_DATUM_PTR(key),
				 greylist_format->dbname,
	                         mu_dbm_strerror());
		rc = 1;
	} else
		rc = 0;

	mu_dbm_close(&db);

	
#line 561
do {
#line 561
  push(env, (STKVAL) (rc));
#line 561
  return;
#line 561
} while (0);
}
}

void
#line 565
bi_db_name(eval_environ_t env)
#line 565

#line 565 "bi_db.m4"
{
#line 565
	
#line 565

#line 565

#line 565
char * fmtid;
#line 565
        
#line 565
get_string_arg(env, 0, &fmtid);
#line 565
        
#line 565

#line 565
        adjust_stack(env, 1);
#line 565

#line 565
	if (prog_trace_option & BUILTIN_MASK_db)
#line 565
		prog_trace(env, "db_name %s",fmtid);;
#line 565

{
	struct db_format *fmt = db_format_lookup(fmtid);
		if (!(fmt != NULL))
#line 568
			env_throw_bi(env, mfe_not_found, "db_name", _("No such db format: %s"),fmtid)
#line 570
;
	do {pushs(env, fmt->dbname);
#line 571
return;
#line 571
} while (0);
}
}

void
#line 575
bi_db_get_active(eval_environ_t env)
#line 575

#line 575 "bi_db.m4"
{
#line 575
	
#line 575

#line 575

#line 575
char * fmtid;
#line 575
        
#line 575
get_string_arg(env, 0, &fmtid);
#line 575
        
#line 575

#line 575
        adjust_stack(env, 1);
#line 575

#line 575
	if (prog_trace_option & BUILTIN_MASK_db)
#line 575
		prog_trace(env, "db_get_active %s",fmtid);;
#line 575

{
	struct db_format *fmt = db_format_lookup(fmtid);
		if (!(fmt != NULL))
#line 578
			env_throw_bi(env, mfe_not_found, "db_get_active", _("No such db format: %s"),fmtid)
#line 580
;
	
#line 581
do {
#line 581
  push(env, (STKVAL) (fmt->enabled));
#line 581
  return;
#line 581
} while (0);
}
}

void
#line 585
bi_db_set_active(eval_environ_t env)
#line 585

#line 585 "bi_db.m4"
{
#line 585
	
#line 585

#line 585

#line 585
char * fmtid;
#line 585
        long  active;
#line 585
        
#line 585
get_string_arg(env, 0, &fmtid);
#line 585
        get_numeric_arg(env, 1, &active);
#line 585
        
#line 585

#line 585
        adjust_stack(env, 2);
#line 585

#line 585
	if (prog_trace_option & BUILTIN_MASK_db)
#line 585
		prog_trace(env, "db_set_active %s %lu",fmtid, active);;
#line 585

{
	struct db_format *fmt = db_format_lookup(fmtid);
		if (!(fmt != NULL))
#line 588
			env_throw_bi(env, mfe_not_found, "db_set_active", _("No such db format: %s"),fmtid)
#line 590
;
	fmt->enabled = active;
}
}

void
#line 595
bi_db_expire_interval(eval_environ_t env)
#line 595

#line 595 "bi_db.m4"
{
#line 595
	
#line 595

#line 595

#line 595
char * fmtid;
#line 595
        
#line 595
get_string_arg(env, 0, &fmtid);
#line 595
        
#line 595

#line 595
        adjust_stack(env, 1);
#line 595

#line 595
	if (prog_trace_option & BUILTIN_MASK_db)
#line 595
		prog_trace(env, "db_expire_interval %s",fmtid);;
#line 595

{
	struct db_format *fmt = db_format_lookup(fmtid);
		if (!(fmt != NULL))
#line 598
			env_throw_bi(env, mfe_not_found, "db_expire_interval", _("No such db format: %s"),fmtid)
#line 600
;
	
#line 601
do {
#line 601
  push(env, (STKVAL) (fmt->expire_interval));
#line 601
  return;
#line 601
} while (0);
}
}


#line 605
void
#line 605
db_init_builtin()
#line 605
{
#line 605
	#line 68 "bi_db.m4"
install_pragma("dbprop", 3, 4, _pragma_dbprop);
#line 196 "bi_db.m4"
va_builtin_install_ex("dbmap", bi_dbmap, 0, dtype_number, 3, 1, 0|0, dtype_string, dtype_string, dtype_number);
#line 206 "bi_db.m4"
va_builtin_install_ex("dbget", bi_dbget, 0, dtype_string, 4, 2, 0|0, dtype_string, dtype_string, dtype_string, dtype_number);
#line 217 "bi_db.m4"
va_builtin_install_ex("dbput", bi_dbput, 0, dtype_unspecified, 5, 2, 0|0, dtype_string, dtype_string, dtype_string, dtype_number, dtype_number);
#line 255 "bi_db.m4"
va_builtin_install_ex("dbdel", bi_dbdel, 0, dtype_unspecified, 4, 2, 0|0, dtype_string, dtype_string, dtype_number, dtype_number);
#line 321 "bi_db.m4"
DBTAB_id = builtin_priv_register(alloc_db_tab, destroy_db_tab);
#line 337 "bi_db.m4"
va_builtin_install_ex("dbfirst", bi_dbfirst, 0, dtype_number, 1, 0, 0|0, dtype_string);
#line 365 "bi_db.m4"
va_builtin_install_ex("dbnext", bi_dbnext, 0, dtype_number, 1, 0, 0|0, dtype_number);
#line 386 "bi_db.m4"
va_builtin_install_ex("dbkey", bi_dbkey, 0, dtype_string, 1, 0, 0|0, dtype_number);
#line 404 "bi_db.m4"
va_builtin_install_ex("dbvalue", bi_dbvalue, 0, dtype_string, 1, 0, 0|0, dtype_number);
#line 463 "bi_db.m4"
	greylist_seconds_left_loc = builtin_variable_install("greylist_seconds_left", dtype_number, VAR_VOLATILE)->off;
#line 470 "bi_db.m4"
va_builtin_install_ex("greylist", bi_greylist, 0, dtype_number, 2, 0, 0|0, dtype_string, dtype_number);
#line 565 "bi_db.m4"
va_builtin_install_ex("db_name", bi_db_name, 0, dtype_string, 1, 0, 0|0, dtype_string);
#line 575 "bi_db.m4"
va_builtin_install_ex("db_get_active", bi_db_get_active, 0, dtype_number, 1, 0, 0|0, dtype_string);
#line 585 "bi_db.m4"
va_builtin_install_ex("db_set_active", bi_db_set_active, 0, dtype_unspecified, 2, 0, 0|0, dtype_string, dtype_number);
#line 595 "bi_db.m4"
va_builtin_install_ex("db_expire_interval", bi_db_expire_interval, 0, dtype_number, 1, 0, 0|0, dtype_string);

#line 605 "bi_db.m4"
		
#line 605
}

#line 603 "../mfd/snarf.m4"

#line 603

